//Values, calculations and variables are based on:
//https://www.waveshare.com/w/upload/9/91/BME280_datasheet.pdf
//and
//https://github.com/BoschSensortec/BME280_driver/blob/master/bme280.c

var dig_T1, dig_T2, dig_T3; //Temperature calibration data
var dig_P1, dig_P2, dig_P3, dig_P4, dig_P5, dig_P6, dig_P7, dig_P8, dig_P9; //Pressure calibration data
var dig_H1, dig_H2, dig_H3, dig_H4, dig_H5, dig_H6; //Humidity calibration data
var t_fine; //Fine resolution temperature value

var i2c; //i2c object
var doRead, connected = false;

function getUnsignedInt16FromBuffer(buffer, index) {
    let arr = new Uint8Array(buffer);
    return (arr[index + 1] << 8) + arr[index];
}

function getSignedInt16FromBuffer(buffer, index) {
    let arr = new Uint8Array(buffer);
    let temp = (arr[index + 1] << 8) + arr[index];
    if (temp > 32767) {
        return temp - 65536;
    }

    return temp;
}

async function readCalibrationData() {
    i2c = new miletus.I2C();
    await i2c.open(0x77);
    connected = true;

    //Set oversampling and sensore mode
    await i2c.writeByte(0xF2, 5); //humidity
    await i2c.writeByte(0xF4, 183); //pressure, temperature, mode

    //Read temperature calibration data
    let buffer = await i2c.readBuffer(0x88, 24);

    dig_T1 = getUnsignedInt16FromBuffer(buffer, 0);
    dig_T2 = getSignedInt16FromBuffer(buffer, 2);
    dig_T3 = getSignedInt16FromBuffer(buffer, 4);

    //Get pressure calibration data
    dig_P1 = getUnsignedInt16FromBuffer(buffer, 6);
    dig_P2 = getSignedInt16FromBuffer(buffer, 8);
    dig_P3 = getSignedInt16FromBuffer(buffer, 10);
    dig_P4 = getSignedInt16FromBuffer(buffer, 12);
    dig_P5 = getSignedInt16FromBuffer(buffer, 14);
    dig_P6 = getSignedInt16FromBuffer(buffer, 16);
    dig_P7 = getSignedInt16FromBuffer(buffer, 18);
    dig_P8 = getSignedInt16FromBuffer(buffer, 20);
    dig_P9 = getSignedInt16FromBuffer(buffer, 22);
  
    //Read humidity calibration data
    dig_H1 = await i2c.readByte(0xA1);
    buffer = await i2c.readBuffer(0xE1, 7);
    dig_H2 = getSignedInt16FromBuffer(buffer, 0);
    let view = new DataView(buffer);
    dig_H3 = view.getUint8(2);
    dig_H4 = (view.getUint8(3) << 4) + (view.getUint8(4) & 0xF);
    dig_H5 = (view.getUint8(5) << 4) + (view.getUint8(4) >>> 4);
    dig_H6 = view.getInt8(6);
}

function compensateTemperature(buffer) {
    let view = new DataView(buffer);
    let data_msb = view.getUint8(0) << 12;
    let data_lsb = view.getUint8(1) << 4;
    let data_xlsb = view.getUint8(2) >>> 4;
    let rawTemp = data_msb | data_lsb | data_xlsb;

    let var1 = (((rawTemp >>> 3) - (dig_T1 << 1)) * dig_T2) >>> 11;
    let var2 = (((((rawTemp >>> 4) - dig_T1) * ((rawTemp >>> 4) - dig_T1)) >>> 12) * dig_T3) >>> 14;

    t_fine = var1 + var2;
    let temp = (t_fine * 5 + 128) >>> 8;

    return temp / 100;
}

function compensatePressure(buffer) {
    let view = new DataView(buffer);
    let data_msb = view.getUint8(0) << 12;
    let data_lsb = view.getUint8(1) << 4;
    let data_xlsb = view.getUint8(2) >>> 4;

    let rawPress = data_msb | data_lsb | data_xlsb;
    let var1 = (t_fine / 2.0) - 64000;
    let var2 = (var1 * var1 * dig_P6) / 32768.0;
    var2 = var2 + (var1 * dig_P5 * 2.0);
    var2 = (var2 / 4.0) + (dig_P4 * 65536.0);
    var1 = (((dig_P3 * var1 * var1) / 524288.0) + (dig_P2 * var1)) / 524288.0;
    var1 = (1.0 + (var1 / 32768.0)) * dig_P1;
    
    if (var1 === 0.0) {
        return 0.0;
    }

    let p = 1048576.0 - rawPress;
    p = ((p - (var2 / 4096.0)) * 6250.0) / var1;
    var1 = (dig_P9 * p * p) / 2147483648.0;
    var2 = (p * dig_P8) / 32768.0;
    p = p + ((var1 + var2 + dig_P7) / 16.0);

    return p / 100;
}

function compensateHumidity(buffer) {
    let view = new DataView(buffer);
    let rawHum = (view.getUint8(0) << 8) + view.getUint8(1);
    let var1 = t_fine - 76800;
    let var2 = rawHum << 14;
    let var3 = dig_H4 << 20;
    let var4 = dig_H5 * var1;
    let var5 = ((var2 - var3 - var4) + 16384) >>> 15;
    var2 = (var1 * dig_H6) >>> 10;
    var3 = (var1 * dig_H3) >>> 11;
    var4 = ((var2 * (var3 + 32768)) >>> 10) + 2097152;
    var2 = ((var4 * dig_H2) + 8192) >>> 14;
    var3 = var5 * var2;
    var4 = ((var3 >>> 15) * (var3 >>> 15)) >>> 7;
    var5 = var3 - ((var4 * dig_H1) >>> 4);

    if (var5 < 0) {
        var5 = 0;
    } else if (var5 > 419430400) {
        var5 = 419430400;
    }

    return (var5 >>> 12) / 1024;
}

async function readSensorData() {
    //Temperature
    let rawValues = await i2c.readBuffer(0xFA, 3);
    let temp = compensateTemperature(rawValues);
    document.getElementById("temperature").innerHTML = temp + ' °C';

    //Pressure
    rawValues = await i2c.readBuffer(0xF7, 3);
    let press = compensatePressure(rawValues);
    document.getElementById("pressure").innerHTML = press.toString().match(/^\d+(?:\.\d{0,2})?/) + ' hPa';

    //Humidity
    rawValues = await i2c.readBuffer(0xFD, 2);
    let hum = compensateHumidity(rawValues);
    document.getElementById("humidity").innerHTML = hum.toString().match(/^\d+(?:\.\d{0,2})?/) + '%';
}

async function startReading() {
    await readCalibrationData();
    document.getElementById("start-reading").disabled = true;
    document.getElementById("stop-reading").disabled = false;
    doRead = true;
    while (doRead) {
        await readSensorData();  
        await new Promise(resolve => setTimeout(resolve, 1000));
    } 
}

function stopReading() {
    doRead = false;
    if (connected) {
        connected = false;
        i2c.close;
    }
    document.getElementById("start-reading").disabled = false;
    document.getElementById("stop-reading").disabled = true;
}

window.onbeforeunload = function () {
    stopReading();
}

document.getElementById("start-reading").onclick = startReading;
document.getElementById("stop-reading").onclick = stopReading;